/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __Kicker_H__
#define __Kicker_H__

#include "Agent.h"

#define STEP_KICK_ANGLE 10 // KickerǶȵĲ360 / 10 = 36Ƕ
#define POINTS_NUM      81 // KickerܵPOINTS_NUM = nlayer[0] + nlayer[1] + nlayer[2]

class Kicker
{
    Kicker();

public:
    ~Kicker();

    static Kicker & instance();

    /**
     * mKickerValue
     * Compute kicker value table.
     */
    void ComputeUtilityTable();

    /**
     * ͨkickģͼmax_rand
     * Calculate maximum random error after kick action.
     */
    double GetMaxRandAfterKick(const Agent & agent, double power);

    /**
     * ĳٶȺĿٶȣõkickҪpowerΪΪ
     * Calculate power need for an kick action according to initial and objective velocity.
     */
    double GetOneKickPower(const Vector & ball_vel, const Vector & ball_goal_vel, const double kick_rate);

    /**
     * 2غڸĿǶȺڵǰ£ʹܴﵽ٣Ϊ
     * The following 2 functions calculate maximum speed the ball can reach by given conditions.
     */
	double GetMaxSpeed(const Agent & agent, double kick_angle, int cycle, Vector * kick_out_pos = NULL);
    double GetMaxSpeed(const Agent & agent, Vector kick_pos, int cycle, Vector * kick_out_pos = NULL);

    /**
     * õĽӿڣҪ
     * the following 2 functions generate kick action plans.
     */
	bool KickBall(Agent & agent, double angle, double speed_out, KickMode mode = KM_Hard, int *cycle_left = NULL, bool is_shoot = false);
    bool KickBall(Agent & agent, const Vector & target, double speed_out, KickMode mode = KM_Hard, int * cycle_left = NULL, bool is_shoot = false);

    /**
     * ִк
     * Execute a kick action plan.
     */
    void Execute(Agent &agent, const ActionPlan &plan);

private:

    /** ɢѰһ */
    inline int NearestPoint(const Vector & p)
    {
        double ang = 0.0;
	    double value = p.Mod();
	    int idxA;
	    if(value < mD1)
	    {
		    double fix = 180.0 / mNlayer[0];
		    ang = p.Dir() + fix;
		    ang = GetNormalizeAngleDeg(ang, 0.0);
		    idxA = int(ang * mNlayer[0] / 360.0);
		    return idxA;
	    }
	    else if(value > mD2)
	    {
		    double fix = 180.0 / mNlayer[2];
		    ang = p.Dir() + fix;
		    ang = GetNormalizeAngleDeg(ang, 0.0);
		    idxA = int(ang * mNlayer[2] / 360.0);
		    return mNlayer[0] + mNlayer[1] + idxA;
	    }
	    else
	    {
		    double fix = 180.0 / mNlayer[1];
		    ang = p.Dir() + fix;
		    ang = GetNormalizeAngleDeg(ang, 0.0);
		    idxA = int(ang * mNlayer[1] / 360.0);
		    return mNlayer[0] + idxA;
	    }
    }

    /** ȡmKickerValue */
    void ReadUtilityTable();

    /** kickݣʱ */
    void UpdateKickData(const Agent & agent);

    /** õһkickĳܴﵽٶȣΪΪ*/
    double GetOneKickMaxSpeed(const Vector & ball_vel, double kick_angle, double max_accel);

    /** ֪󣬸ú */
    bool KickBall(Agent & agent, const Vector & target, double speed_out, int cycle, bool is_shoot = false);

    /** ǵõActionInfoӿڵĺ */
	AtomicAction GetOneKickAction(const Agent & agent, const Vector & ball_pos, const Vector & ball_vel, const Vector & target, bool is_self = true);
	AtomicAction GetKickOutAction(const Agent & agent, Vector & ball_pos, Vector & ball_vel, bool is_self = true);
    AtomicAction GetMaxKickOutAction(const Agent & agent, Vector & ball_pos, Vector & ball_vel, bool is_self = true);
    AtomicAction GetTurnAction(const Agent & agent);

    /** һͶԼתĹ滮 */
    ActionPlan OneCycleKick(const Agent & agent, bool is_shoot = false);
    ActionPlan MultiCycleKick(const Agent & agent, int cycle);
    AtomicAction TurnPlan(const Agent & agent, int index, double turn_max_speed);

private:
    AgentID     mAgentID;

    int         mNlayer[3];         /** ÿɢ */
    double      mD1;                /** 1,2뾶ľֵ */
    double      mD2;                /** 2,3뾶ľֵ */
    double      mDlayer[3];         /** ÿİ뾶 */
    Vector      mPoint[POINTS_NUM]; /** 洢е */
    float       mKickerValue[3][36][POINTS_NUM][POINTS_NUM]; /** 2,3,436ǶȣPOINTS_NUM */ // floatɣʡռռ

    ReciprocalCurve mOppCurve;      /** ֵӰ */
    ReciprocalCurve mRandCurve;     /** Ӱ */
    ReciprocalCurve mSpeedCurve;    /** ٵӰ */

    ActionInput mInput;                 /** ¼kick״̬Ϣ */
    double      mKickSpeed;             /** kickٶȴС */
    Vector      mKickTarget;            /** kickĿ㣬playerϵ */
    //AngleDeg	mKickAngle;             /** kickٶȷ巽-180180֮ */

    double      mKickRate[POINTS_NUM];  /** ÿԱkick_rate */
    double      mMaxAccel[POINTS_NUM];  /** ÿԱܲٶeff_power */
    double      mRandEva[POINTS_NUM];   /** ÿԱkick */
    double      mPointEva[POINTS_NUM];  /** ÿԱkickeva */

    double      mMaxRandFactor;         /** ¼kick randʱһ */

	AgentID     mMaxSpeedFlag[360][3];  /** ¼Ƕȵļǣ */
    double      mMaxSpeed[360][3];      /** 360Ƕȣ1,2,3ܴﵽٶ */
    Vector      mMaxSpeedOutPos[360][3];/** 360Ƕȣ1,2,3´ﵽٶʱĳλ */

public:
    /** ṩйagentĽӿڣֱӷAtomicActionⲿ */

    /**
     * ֤kickʹٴﵽС
     * Get an atomic action to stop ball or minimum ball speed.
     */
    AtomicAction GetStopBallAction(const Agent & agent, bool forceStop = true);

    /**
     * ٵضĿٶ
     * Accelerate the ball to a certain velocity.
     */
    AtomicAction GetAccelerateBallAction(const Agent & agent, Vector & goal_vel);

    /**
     * ض
     * Kick the ball to a certain direction.
     */
    AtomicAction GetKickBallToAngleAction(const Agent & agent, AngleDeg rel_angle, double rel_dist_rate = 0.9);
};


#endif

